from ..nodes import CompositorNodeGroup
import bpy


class DepthVis(CompositorNodeGroup):
    def __init__(self, node_tree, max_depth=1, col=(1, 1, 1)):
        super().__init__(name="DepthVis", node_tree=node_tree)
        self.max_depth = 1

        self.add_socket("NodeSocketFloat", "Depth", "INPUT")
        self.add_socket("NodeSocketColor", "Image", "OUTPUT")

        self.map_range_node = self.group.nodes.new("CompositorNodeMapRange")
        self.map_range_node.inputs["From Max"].default_value = max_depth

        self.rgb_node = self.group.nodes.new("CompositorNodeRGB")
        if len(col) == 3:
            col = (*col, 1)

        self.rgb_node.outputs["RGBA"].default_value = col

        self.multiple_node = self.group.nodes.new("CompositorNodeMath")
        self.multiple_node.operation = "MULTIPLY"

        self.group.links.new(
            self.input_node.outputs["Depth"], self.map_range_node.inputs["Value"]
        )
        self.group.links.new(
            self.map_range_node.outputs["Value"], self.multiple_node.inputs[0]
        )
        self.group.links.new(
            self.rgb_node.outputs["RGBA"], self.multiple_node.inputs[1]
        )
        self.group.links.new(
            self.multiple_node.outputs["Value"], self.output_node.inputs["Image"]
        )

        self.tidy()



import bpy
from .utils import project_points
import mathutils
import numpy as np
from ..blender.mesh import Mesh
from typing import List, Union
from .annotation_handler import AnnotationHandler
from ..blender.camera import Camera


def _get_object_axes(
    object: Mesh,
    camera: bpy.types.Camera = None,
    scene: bpy.types.Scene = None,
    normalized=True,
    invert_y=True,
) -> np.ndarray:
    """Get the unit axes of an object in camera space.

    :param object: Mesh
    :param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
    :param scene: bpy.types.Scene (if None, use bpy.context.scene)
    :param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)

    :return: (4 x 2) centroid + 3 pixel positions of ends of unit axes [XYZ] in XY
    """

    if camera is None:
        camera = bpy.context.scene.camera

    if scene is None:
        scene = bpy.context.scene

    if len(object._meshes) > 1:
        raise NotImplementedError("Axes only supported for single meshes")

    centroid = np.array(object.centroid())
    verts = np.concatenate([centroid[None, :], centroid[None, :] + object.axes], axis=0)
    coords_2d = project_points(verts, scene, camera, invert_y=invert_y)

    if normalized:
        coords_2d[:, 0] /= scene.render.resolution_x
        coords_2d[:, 1] /= scene.render.resolution_y

    return coords_2d


def _get_objects_axes(
    objects: List[Mesh],
    camera: bpy.types.Camera = None,
    scene: bpy.types.Scene = None,
    normalized: bool = False,
    invert_y: bool = True,
) -> List[np.ndarray]:
    """Run `get_axes` for multiple objects

    :param objects: List of Mesh objects
    :param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
    :param scene: bpy.types.Scene (if None, use bpy.context.scene)
    :param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
    """

    return [
        _get_object_axes(
            obj, camera=camera, scene=scene, normalized=normalized, invert_y=invert_y
        )
        for obj in objects
    ]


def get_axes(
    objects: List[Mesh],
    camera: Union[Camera, List[Camera]] = None,
    scene: bpy.types.Scene = None,
    normalized: bool = False,
    invert_y: bool = True,
) -> AnnotationHandler:
    """Project local axes for all objects into all cameras

    :param objects: List of Mesh objects
    :param camera: Either a single :class:`~blendersynth.blender.camera.Camera` object, or a list of them. If None, use bpy.context.scene.camera
    :param scene: bpy.types.Scene (if None, use bpy.context.scene)
    :param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
    """

    return_dict = {}

    if camera is None:
        camera = bpy.context.scene.camera

    if not isinstance(camera, list):
        camera = [camera]

    for cam in camera:
        return_dict[cam.name] = _get_objects_axes(
            objects, camera=cam, scene=scene, normalized=normalized, invert_y=invert_y
        )

    return AnnotationHandler.from_annotations(return_dict, ann_type="axes")



import os
import bpy


def load_image(pth: str):
    """Load an image from a path. Use :func:`os.path.abspath` to avoid rel path issues on Windows.

    :param pth: Path to the image

    """
    abs_pth = os.path.abspath(pth)
    return bpy.data.images.load(abs_pth)



import bpy
from .bsyn_object import BsynObject

prims = {
    "circle": bpy.ops.curve.primitive_bezier_circle_add,
    "bezier": bpy.ops.curve.primitive_bezier_curve_add,
    "path": bpy.ops.curve.primitive_nurbs_path_add,
}


class Curve(BsynObject):
    """Blender path object"""

    def __init__(
        self, path_type="circle", scale=1, location=(0, 0, 0), rotation=(0, 0, 0)
    ):
        """Create a new path:

        :param path_type: Type of path to create. One of 'circle', 'bezier', 'path' (straight line)
        """

        self.path = None
        self._create_path(path_type)
        self._object = self.path

        self.scale = scale
        self.location = location
        self.rotation = rotation

    def _create_path(self, path_type):
        prims[path_type]()
        self.path = bpy.context.object



from collections import namedtuple
from copy import deepcopy
from ..utils import types

# docs-special-members: __init__

ANN_TYPES = ["bbox", "keypoints", "axes"]
"""Valid annotation types"""


class Annotation:
    """A class for storing annotations for a single image."""

    def __init__(
        self,
        bbox: types.BboxAnnotation = None,
        keypoints: types.KeypointOrAxesAnnotation = None,
        axes: types.KeypointOrAxesAnnotation = None,
    ):
        """

        :param bbox: Bounding box tuple per instance
        :param keypoints: [N x 2] array per instance
        :param axes: [4 x 2] array per instance
        """
        self.bbox = bbox
        self.keypoints = keypoints
        self.axes = axes

    def __getitem__(self, item):
        return getattr(self, item)

    def _set(self, item, value) -> "Annotation":
        """In place set, returning self"""
        setattr(self, item, value)
        return self

    def set(self, item: str, value, inplace: bool = False) -> "Annotation":
        """Return a new Annotation, with item set to value.

        :param item: A key in :class:`ANN_TYPES`
        :param value: Value to set
        :param inplace: If True, set in place and return self
        """
        if inplace:
            return self._set(item, value)

        return self.copy()._set(item, value)

    def copy(self) -> "Annotation":
        """Return a deep copy of this Annotation instance"""
        return deepcopy(self)

    def union(self, other: "Annotation", priority="other") -> "Annotation":
        """Joins two annotations together, returning a new Annotation instance.

        :param other: Annotation to join with
        :param priority: Which annotation to take priority when merging
        """

        combined = Annotation()

        low_priority = "self" if priority == "other" else "other"

        for ann_type in ANN_TYPES:
            vals = {"self": self[ann_type], "other": other[ann_type]}

            if vals[priority] is not None:
                combined._set(ann_type, vals[priority])

            elif vals[low_priority] is not None:
                combined._set(ann_type, vals[low_priority])

        return combined

    def __add__(self, other: "Annotation"):
        return self.union(other)


class AnnotationHandler:
    """An object for storing and handling per-view :class:`Annotation` instances"""

    def __init__(self, data: dict):
        """
        :param data: Dictionary of camera_name: Annotation class
        """

        self._data = data

    @classmethod
    def from_annotations(
        cls, data: dict, ann_type: str = "bbox"
    ) -> "AnnotationHandler":
        """Create an Annotation Handler from a single type of annotation

        :param data: Dictionary of camera_name: annotation (e.g. bboxes)
        :param ann_type: Type of annotation (e.g. 'bbox')
        """

        assert (
            ann_type in ANN_TYPES
        ), f"Invalid annotation type: {ann_type}. Must be one of {ANN_TYPES}"
        out_data = {}

        for cam_name, v in data.items():
            ann = Annotation(**{ann_type: v})
            out_data[cam_name] = ann

        return cls(out_data)

    def union(
        self, other: "AnnotationHandler", priority: str = "other"
    ) -> "AnnotationHandler":
        """Combine two AnnotationHandlers, merging annotations wherever possible. Returns a new instance
        comprising the union of the two AnnotationHandlers.

        :param other: AnnotationHandler to merge with
        :param priority: Which AnnotationHandler to take priority when merging
        """

        assert isinstance(
            other, AnnotationHandler
        ), f"other must be an AnnotationHandler, not {type(other)}"

        out_data = {}
        camera_names = set(self._data.keys()).union(set(other._data.keys()))

        for camera_name in camera_names:
            self_ann, other_ann = self.get_annotation_by_camera(
                camera_name
            ), other.get_annotation_by_camera(camera_name)
            out_data[camera_name] = self_ann.union(other_ann, priority=priority)

        return AnnotationHandler(out_data)

    def get_annotation_by_camera(self, camera_name) -> Annotation:
        """Return annotation for a given camera (default to an empty :class:`Annotation` instance if not found)"""
        return self._data.get(camera_name, Annotation())

    def __add__(self, other: "AnnotationHandler"):
        return self.union(other)



"""Armature object for managing pose"""
import numpy as np

from .bsyn_object import BsynObject, animatable_property
from .other_objects import Empty
from .utils import SelectObjects, SetMode
from typing import Union
from ..utils import types
import bpy
import mathutils


class PoseBone(BsynObject):
    def __init__(self, object: bpy.types.PoseBone):
        self._object = object
        object.rotation_mode = "XYZ"  # all rotations done in XYZ Euler mode

    @property
    def name(self) -> str:
        return self.obj.name

    @property
    def constraints(self):
        return self.obj.constraints

    @property
    def armature(self):
        return self.obj.id_data

    @property
    def matrix_world(self):
        return self.armature.matrix_world @ self.obj.matrix

    @property
    def location(self) -> mathutils.Vector:
        """
        Gets the world location of the head (start) of the bone.
        """
        return self.obj.head

    @property
    def head_location(self) -> mathutils.Vector:
        """
        Gets the world location of the head (start) of the bone.
        """
        return self.armature.matrix_world @ self.obj.head

    @property
    def tail_location(self) -> mathutils.Vector:
        """
        Gets the world location of the tail (or end) of the bone.
        """
        return self.armature.matrix_world @ self.obj.tail

    @property
    def rotation_euler(self) -> mathutils.Euler:
        """Rotation of pose bone in world space"""
        return self.matrix_world.to_euler()

    @property
    def local_rotation_euler(self) -> mathutils.Euler:
        """Rotation of pose bone in local space (ignoring parent transforms)"""
        return self.obj.rotation_euler

    @local_rotation_euler.setter
    def local_rotation_euler(self, value: mathutils.Euler):
        self.obj.rotation_euler = value

    def apply_global_rotation(self, axis: str = "X", val: float = 0.0, degrees=False):
        """Equivalent to selecting the bone in Blender, and applying a global rotation about an axis

        :param axis: X, Y, or Z
        :param val: Rotation amount in radians
        :param degrees: If true, rotation is given in degrees"""

        if degrees:
            val = np.deg2rad(val)

        with SetMode("POSE", object=self.armature):
            with SelectObjects([self.obj]):
                bpy.context.object.data.bones.active = self.armature.data.bones[
                    self.name
                ]

                bpy.ops.transform.rotate(
                    value=val,
                    orient_axis=axis,
                    orient_type="GLOBAL",
                    constraint_axis=[axis == "X", axis == "Y", axis == "Z"],
                )

    @animatable_property("scale")
    def set_scale(self, scale: types.VectorLikeOrScalar, update: bool = True):
        """Set scale of pose bone.

        :param scale: Scale to set. Either single value or 3 long vector
        :param update: Update the scene after setting
        scale. For batch operations, can be faster to set this to False, and call `bsyn.context.view_layer.update()`
        after all operations.
        """

        if isinstance(scale, (int, float)):
            scale = (scale, scale, scale)

        self.obj.scale = scale

        if update:
            bpy.context.view_layer.update()


class BoneConstraint(BsynObject):
    """Generic pose bone constraint object"""

    def __init__(
        self, pose_bone: PoseBone, constraint: bpy.types.Constraint, empty: Empty
    ):
        self.pose_bone = pose_bone
        self.constraint = constraint
        self.empty = empty
        self._object = empty.obj

    @property
    def name(self):
        return f"{self.constraint.name}: {self.pose_bone.name} -> {self.empty.name}"

    @property
    def bone(self):
        return self.pose_bone

    def remove(self):
        """Remove this constraint."""
        self.bone.constraints.remove(self.constraint)
        if self.empty is not None:
            self.empty.remove()

    @property
    def armature(self):
        return self.bone.id_data


class Armature(BsynObject):
    """This class manages armatures - a collection of posable bones."""

    def __init__(self, object: bpy.types.Armature):
        self._object = object
        self.ik_constraints = {}  # for IK constraints
        self.constraints = {}  # for generic constraints
        self.pose_bones = {n: PoseBone(b) for n, b in object.pose.bones.items()}

    @property
    def name(self) -> str:
        return self.obj.name

    @property
    def pose(self) -> bpy.types.Pose:
        return self.obj.pose

    def get_bone(self, bone: Union[str, PoseBone, bpy.types.PoseBone]) -> PoseBone:
        """
        Get bone from armature.

        :param bone_name: Name of bone to get (or PoseBone object)
        :return: PoseBone object
        """

        if isinstance(bone, PoseBone):
            return bone

        if isinstance(bone, bpy.types.PoseBone):
            return self.pose_bones[bone.name]

        if isinstance(bone, str):
            return self.pose_bones[bone]

        raise TypeError(
            f"Expected bone to be PoseBone, str, or PoseBone, got {type(bone)}"
        )

    def pose_bone(
        self,
        bone: Union[str, bpy.types.PoseBone],
        rotation: types.VectorLike = None,
        location: types.VectorLike = None,
        scale: types.VectorLike = None,
        frame: int = None,
    ):
        """Set the pose of a bone by giving a Euler XYZ rotation and/or location.

        :param bone: Name of bone to pose, or PoseBone object
        :param rotation: Euler XYZ rotation in radians
        :param location: Location of bone
        :param scale: Scale of bone
        :param frame: Frame to set pose on. If given, will insert keyframe here.
        """

        with SetMode("POSE", object=self.obj):
            if isinstance(bone, str):
                bone = self.get_bone(bone)

            bone.rotation_mode = "XYZ"
            if rotation is not None:
                bone.set_rotation_euler(rotation, frame=frame)

            if location is not None:
                bone.set_location(location, frame=frame)

            if scale is not None:
                bone.set_scale(scale, frame=frame)

    def clear_pose(self, rot=True, location=True, scale=True, bones=None):
        """Clear the pose of the armature.
        For the target bones, sets poses to zero, and removes any IK constraints.

        :param rot: Clear rotation
        :param location: Clear location
        :param scale: Clear scale
        :param bones: List of bones to clear. If not given, will clear all bones.
        """

        with SetMode("POSE", object=self.obj):
            for bone in self.pose.bones:
                if bones is None or bone.name in bones:
                    if rot:
                        bone.rotation_euler = (0, 0, 0)
                    if location:
                        bone.location = (0, 0, 0)
                    if scale:
                        bone.scale = (1, 1, 1)

                for bone_name in self.ik_constraints.keys():
                    if bone_name == bone.name:
                        self.ik_constraints[bone_name].remove()  # remove the constraint
                        self.ik_constraints.pop(bone_name)  # remove reference to it

                for cname in [*self.constraints.keys()]:
                    constraint = self.constraints[cname]
                    if constraint.pose_bone.name == bone.name:
                        self.constraints[cname].remove()
                        self.constraints.pop(cname)

    def add_constraint(
        self,
        bone: Union[str, PoseBone],
        constraint_name: str,
        target: Empty = None,
        **kwargs,
    ) -> BoneConstraint:
        """
        Applies a generic bone constraint. Returns BoneConstraint object

        :param bone: Bone to apply constraint to
        :param constraint_name:
        :param target: Empty target. If none given, will create one.

        Any other constraint specific keyword arguments given as well will be fed into the new constraint
        """

        bone = self.get_bone(bone)

        # Add IK constraint to the specified bone
        constraint = bone.constraints.new(constraint_name)
        for k, v in kwargs.items():
            setattr(constraint, k, v)

        # Create an empty to serve as the IK target
        if target is None:
            target = Empty(name="Target_For_" + bone.name)

        constraint.target = target.object

        bpy.context.view_layer.update()

        constraint = BoneConstraint(bone, constraint, target)
        self.constraints[constraint.name] = constraint

        return constraint



"""Quick utility to run the current script from Blender"""
import subprocess
import inspect
import sys
import os
from ..utils.blender_setup.blender_locator import get_blender_path
from ..file.tempfiles import create_temp_file, cleanup_temp_files as cleanup
from shutil import copyfile


def _copy_over_script(filepath: str) -> str:
    """Copies over a python script to a tempfile, returning the path.
    Removes certain lines so it can run in blender"""
    remove_lines_containing = [".run_this_script"]
    new_lines = []
    with open(filepath, "r") as f:
        lines = f.readlines()
        for line in lines:
            if not any([s in line for s in remove_lines_containing]):
                new_lines.append(line)

    new_filepath = create_temp_file(ext=".py")
    with open(new_filepath, "w") as f:
        f.writelines(new_lines)

    return new_filepath


def is_blender_running():
    """Returns True if blender is running, False otherwise"""
    try:
        import bpy

        return bpy.__file__ is not None
    except:
        return False


def run_this_script(
    *args,
    open_blender: bool = False,
    debug: bool = False,
    IDE: str = "PyCharm",
    port: int = 5678,
    host: str = "localhost",
    blend_src: str = None,
    blend_as_copy: bool = False,
    **kwargs,
):
    """Run the script in which this function is called from Blender.

    :param open_blender: If True, open a Blender instance after all code is executed, otherwise run in background
    :param debug: If True, will run in debug mode
    :param IDE: IDE to use for debugging. Currently only PyCharm and VSCode are supported
    :param port: Port to use for debugging
    :param host: Host to use for debugging
    :param blend_src: Path to blend file to open (note: this is preferable to `blendersynth.load_blend` as it handles context better)
    :param blend_as_copy: If True, will copy `blend_src` to a temp file before opening - this is useful if you want to make sure you don't accidentally override your `blend_src` file

    args & kwargs are passed to the script being run as command line arguments.

    The flag `--run_this_script` is passed to the script being run to indicate that it is being run from `run_this_script`.
    Use function `is_from_run_this_script` to check if the script is being run from `run_this_script`.
    """
    running_in_blender = is_blender_running()

    caller_path = inspect.stack()[
        1
    ].filename  # path of script that called this function

    if not running_in_blender:  # if blender is not running this script
        caller_dir = os.path.dirname(caller_path)
        env = os.environ.copy()
        env["PYTHONPATH"] = caller_dir + os.pathsep + env.get("PYTHONPATH", "")

        blender_path = get_blender_path()

        if blend_src is not None and blend_as_copy:
            new_filepath = create_temp_file(ext=".blend")
            copyfile(blend_src, new_filepath)
            blend_src = new_filepath

        commands = (
            [blender_path]
            + [blend_src] * (blend_src is not None)
            + ["--background"] * (not open_blender)
            + ["--python", caller_path, "--"]
        )

        for arg in args:
            commands += [f"--{arg}"]

        for key, value in kwargs.items():
            commands += [f"--{key}", str(value)]

        commands += ['--run_this_script']  # flag to indicate that this is being run from run_this_script

        subprocess.call(commands, env=env)

        cleanup()  # cleanup temp files
        sys.exit()  # exit the script once blender is finished

    else:
        if debug:
            IDE = IDE.lower()
            assert IDE in [
                "pycharm",
                "vscode",
            ], f"IDE `{IDE}` not supported, only PyCharm and VSCode are supported"

            if IDE == "pycharm":
                import pydevd_pycharm

                pydevd_pycharm.settrace(
                    host,
                    port=port,
                    stdoutToServer=True,
                    stderrToServer=True,
                    suspend=False,
                )

            elif IDE == "vscode":
                import debugpy

                debugpy.listen((host, port))
                debugpy.wait_for_client()

        # blender is running this script
        if open_blender:
            # load the script into blender for viewing
            import bpy

            # REMOVED FUNCTIONALITY TO LOAD SCRIPT TEXT INTO BLENDER AS CAUSING ISSUES
            # from ..utils import layout

            # caller_path = bpy.path.abspath(caller_path)
            # script_path = _copy_over_script(caller_path)

            # text_block = bpy.data.texts.load(script_path)
            # layout.change_area_to("DOPESHEET_EDITOR", "TEXT_EDITOR")
            # layout.get_area("TEXT_EDITOR").spaces[0].text = text_block

def is_from_run_this_script():
    """Returns True if this script is being run from run_this_script"""
    return '--run_this_script' in sys.argv


# First, make sure Blender has been set up properly
import sys
from .utils.blender_setup import check_blender_install, install_module


def fix_blender_install(local=False, editable=False):
    check_blender_install(
        force_all=True, blendersynth_from_local=local, blendersynth_editable=editable
    )


def fix_blender_modules(local=False, editable=False):
    check_blender_install(
        force_install_dependencies=True,
        blendersynth_from_local=local,
        blendersynth_editable=editable,
    )


from .run.run_this_script import run_this_script, is_blender_running, is_from_run_this_script

from .utils.blender_setup.blender_locator import get_blender_path
from .run.import_handling import conditional_import, IllegalImport
from typing import TYPE_CHECKING

is_building_docs = "sphinx" in sys.modules
is_blender = is_blender_running()
BLENDER_IMPORTS = (
    is_building_docs or is_blender
)  # if blender is running this script, or if building docs

if BLENDER_IMPORTS or TYPE_CHECKING:
    from .blender.mesh import Mesh
    from .blender.material import Material
    from .blender.curve import Curve
    from .blender.other_objects import Empty
    from .blender import render
    from .blender.compositor.compositor import Compositor
    from .blender import aov
    from .file.dataset_inputs import Inputs, DebugInputs
    from .file.tempfiles import cleanup_temp_files
    from .run.blender_interface import log_event
    from .blender.world import world
    from .blender.light import Light
    from .blender.camera import Camera
    from . import annotations
    from .utils import layout
    from .blender import utils as blender_utils

    import mathutils
    import bpy

    # set IllegalImports here to warn users if they access the objects in the wrong python environment
    execute_jobs, install_module = [
        IllegalImport(REQUIRES_BLENDER=False) for _ in range(2)
    ]

else:
    # Vanilla only imports
    from .run.run import execute_jobs
    from .utils.blender_setup.check_blender_install import install_module

    # set IllegalImports here to warn users if they access the objects in the wrong python environment
    (
        Mesh,
        Material,
        Curve,
        Empty,
        render,
        Compositor,
        aov,
        Inputs,
        DebugInputs,
        cleanup_temp_files,
        log_event,
        world,
        Light,
        Camera,
        annotations,
        layout,
        blender_utils,
        *___,
    ) = [IllegalImport() for _ in range(99)]


# Global imports
from . import file

# scripting setup
if is_blender:
    from bpy import *  # import everything from bpy

    def load_blend(src):
        return bpy.ops.wm.open_mainfile(filepath=src)

    # set render engine to cycles
    from .run.pre_ops import on_script_open
    from .run.load_blend import load_blend

    has_loaded_scene = len(sys.argv) > 1 and ".blend" in sys.argv[1]
    on_script_open(delete_existing=not has_loaded_scene)

if not (BLENDER_IMPORTS or TYPE_CHECKING):
    check_blender_install(
        blendersynth_from_local="--local" in sys.argv
    )  # check install here



import bpy
from .nodes.node_arranger import tidy_tree
from typing import Union
from ..utils import io


class World:
    """World object - for managing world nodes, HDRIs, etc."""

    def __init__(self):
        self.mode = "Color"
        self.world = bpy.data.worlds["World"]
        self.node_tree = self.world.node_tree
        self.world_nodes = self.node_tree.nodes

        self.hdri_link = None

        # if default setup (a Background node and a WorldOutput node)
        if len(self.world_nodes) == 2 and all(
            [node.name in ["Background", "World Output"] for node in self.world_nodes]
        ):
            self._setup_nodes()
            self.set_color((0.051,) * 3)  # match blender's default world color

        else:
            self._load_nodes()
            self._setup_hdri()

    def _check_exists(self):
        """Node tree may have been wiped, eg by loading a new file.
        Check if the nodes exist, and if not, recreate them."""

        try:
            self.node_tree.rna_type
        except ReferenceError:
            self.__init__()

    def _setup_nodes(self):
        self._check_exists()
        nodes = self.world_nodes
        nodes.clear()

        self.node_texture = nodes.new(type="ShaderNodeTexEnvironment")
        self.node_background = nodes.new(type="ShaderNodeBackground")
        self.node_output = nodes.new(type="ShaderNodeOutputWorld")

        self.node_tree.links.new(
            self.node_background.outputs["Background"],
            self.node_output.inputs["Surface"],
        )

        tidy_tree(self.node_tree)

    def _load_nodes(self):
        """If Nodes already exist, load them"""
        try:
            self.node_texture = self.world_nodes["Environment Texture"]
            self.node_background = self.world_nodes["Background"]
            self.node_output = self.world_nodes["World Output"]
        except KeyError:
            raise KeyError("To load in a World, the World must have 'Environment Texture', 'Background', and 'World Output' nodes.")

        # add links if they don't exist
        if len(self.node_tree.links) == 0:
            self.node_tree.links.new(
                self.node_background.outputs["Background"],
                self.node_output.inputs["Surface"],
            )


    def _setup_color(self):
        self._check_exists()
        if self.mode == "Color":
            return

        if self.hdri_link is not None:
            self.node_tree.links.remove(self.hdri_link)

        tidy_tree(self.node_tree)

        self.mode = "Color"

    def _setup_hdri(self):
        self._check_exists()
        if self.mode == "HDRI":
            return

        # Link the nodes
        self.hdri_link = self.node_tree.links.new(
            self.node_texture.outputs["Color"], self.node_background.inputs["Color"]
        )
        tidy_tree(self.node_tree)

        self.mode = "HDRI"

    def set_color(self, color: Union[list, tuple], affect_scene: bool = True):
        """Set the world color.

        :param color: RGB or RGBA color
        :param affect_scene: Toggle for whether color's lighting should affect the scene (if False, functions as a solid background color)
        """

        self._setup_color()

        assert len(color) in [3, 4], "Color must be RGB or RGBA"

        if len(color) == 3:
            color = (*color, 1.0)

        self.node_background.inputs["Color"].default_value = color
        self._lighting_from_background(affect_scene)

    def set_hdri(self, pth: str, affect_scene: bool = True, intensity: float = None):
        """Set the HDRI image location

        :param pth: Path to the HDRI image (.hdr or .exr)
        :param affect_scene: Toggle for whether color's lighting should affect the scene (if False, functions as a solid background color)
        :param intensity: [Optional] Set the intensity of the HDRI lighting"""

        self._setup_hdri()
        self.world_nodes["Environment Texture"].image = io.load_image(pth)
        self._lighting_from_background(affect_scene)

        if intensity:
            self.set_intensity(intensity)

    def set_intensity(self, intensity: float = 1.0):
        """Set the intensity of the color/HDRI.

        :param intensity: The intensity value
        """
        self.world_nodes["Background"].inputs[1].default_value = intensity

    def set_transparent(self, transparent=True):
        bpy.context.scene.render.film_transparent = transparent

    def _lighting_from_background(self, val=True):
        """Change the ability for the background to influence lighting on the scene"""
        self.world.cycles_visibility.diffuse = val


world = World()



"""Custom node groups"""
import bpy
from .node_arranger import tidy_tree
from ...utils import version

import numpy as np
import os
import cv2

# docs-special-members: __init__
# no-inherited-members


def tree_add_socket(tree: bpy.types.NodeTree, socket_type: str, name: str, in_out: str):
    """Create a new socket, compatible with Blender <4

    :param tree: NodeTree to add socket to
    :param socket_type: Type of socket
    :param name: Name of socket
    :param in_out: INPUT or OUTPUT"""

    assert in_out in ["INPUT", "OUTPUT"], f"Invalid in_out: {in_out}"

    if version.is_version_plus(4):
        tree.interface.new_socket(socket_type=socket_type, name=name, in_out=in_out)

    else:
        if in_out == "INPUT":
            tree.inputs.new(socket_type, name)
        else:
            tree.outputs.new(socket_type, name)


class NodeGroup:
    """Generic Node Group"""

    TYPE = "Compositor"

    def __init__(self, name: str, node_tree: bpy.types.NodeTree):
        """
        A generic NodeGroup class

        :param name: Name of node group
        :param node_tree: NodeTree to add group to
        """
        self.name = name
        self.node_tree = node_tree
        self.group = bpy.data.node_groups.new(type=f"{self.TYPE}NodeTree", name=name)

        self.gn = group_node = node_tree.nodes.new(f"{self.TYPE}NodeGroup")
        group_node.node_tree = self.group

        self.input_node = self.group.nodes.new("NodeGroupInput")
        self.output_node = self.group.nodes.new("NodeGroupOutput")

    def tidy(self):
        tidy_tree(self.group)

    @property
    def inputs(self) -> dict:
        """Input sockets"""
        return self.gn.inputs

    @property
    def outputs(self) -> dict:
        """Output sockets"""
        return self.gn.outputs

    def input(self, name: str) -> bpy.types.NodeSocket:
        """Get input socket by name"""
        return self.inputs[name]

    def output(self, name: str) -> bpy.types.NodeSocket:
        """Get output socket by name"""
        return self.outputs[name]

    def add_node(self, key: str) -> bpy.types.Node:
        """Create a new node in the group by name"""
        return self.group.nodes.new(key)

    def add_socket(self, socket_type: str, name: str, in_out: str):
        """
        Create a new socket, compatible with Blender <4

        :param socket_type: Type of socket
        :param name: Name of socket
        :param in_out: INPUT or OUTPUT
        """
        tree_add_socket(self.group, socket_type, name, in_out)

    def link(
        self, from_socket: bpy.types.NodeSocket, to_socket: bpy.types.NodeSocket
    ) -> bpy.types.NodeLink:
        """
        Link two sockets in the group

        :param from_socket: Socket to link from
        :param to_socket: Socket to link to
        """
        return self.group.links.new(from_socket, to_socket)

    def __str__(self):
        return f"{self.TYPE}NodeGroup({self.name})"

    def update(self, camera=None, scene=None):
        pass

    def save_image(self, loc: str, image_data: np.ndarray):
        """Save an image to a location. Will also reload any node reference to the image.

        :param loc: Location to save image to"""
        cv2.imwrite(loc, image_data)
        fname = os.path.basename(loc)
        if fname in bpy.data.images:
            bpy.data.images[fname].reload()


class CompositorNodeGroup(NodeGroup):
    """Node Group for use in the compositor"""

    TYPE = "Compositor"


class ShaderNodeGroup(NodeGroup):
    """Node Group for use in the shader editor"""

    TYPE = "Shader"



import bpy
from typing import Union

def is_version(version_num: Union[int, float]) -> bool:
    """Check if the Blender version is equal to the given version number.

    If integer provided, only checks major version.

    :param version_num: version number to check against
    """

    major, minor = map(int, bpy.app.version_string.split(".")[:2])

    if isinstance(version_num, int):
        return major == version_num

    target_major, target_minor = map(int, str(version_num).split("."))

    return (major, minor) == (target_major, target_minor)


def is_version_plus(version_num: Union[int, float]) -> bool:
    """Check if the Blender version is greater than or equal to the given version number.

    is_version_plus(4.0) equivalent to is version 4.0+

    :param version_num: version number to check against
    """

    version_num = float(version_num)
    major, minor = map(int, bpy.app.version_string.split(".")[:2])
    target_major, target_minor = map(int, str(version_num).split("."))
    return (major, minor) >= (target_major, target_minor)



import bpy
from ..utils import get_node_by_name
import os
import shutil
from ..render import render, render_depth
from ..nodes import CompositorNodeGroup
from ..aov import AOV
from .mask_overlay import MaskOverlay
from .visuals import DepthVis
from .image_overlay import (
    KeypointsOverlay,
    BoundingBoxOverlay,
    AlphaImageOverlay,
    AxesOverlay,
)
from ...annotations import Annotation, AnnotationHandler
from ..nodes import tidy_tree
from ..world import world
from ..camera import Camera
from ...utils import version

from typing import Union, List

# Mapping of file formats to extensions
format_to_extension = {
    "BMP": ".bmp",
    "IRIS": ".rgb",
    "PNG": ".png",
    "JPEG": ".jpg",
    "JPEG2000": ".jp2",
    "TARGA": ".tga",
    "TARGA_RAW": ".tga",
    "CINEON": ".cin",
    "DPX": ".dpx",
    "OPEN_EXR_MULTILAYER": ".exr",
    "OPEN_EXR": ".exr",
    "HDR": ".hdr",
    "TIFF": ".tif",
    # Add more formats if needed
}

AVAILABLE_FORMATS = [
    "BMP",
    "IRIS",
    "PNG",
    "JPEG",
    "JPEG2000",
    "TARGA",
    "TARGA_RAW",
    "CINEON",
    "DPX",
    "OPEN_EXR_MULTILAYER",
    "OPEN_EXR",
    "HDR",
    "TIFF",
]
"""List of available output file formats"""

# docs-special-members: __init__


def _default_color_space():
    """Get default color space for Blender version"""
    if version.is_version_plus(4):
        return "AgX Base sRGB"
    else:
        return "Filmic sRGB"


def _get_badfname(fname, N=100):
    """Search for filename in the format
    <main_fname><i:04d>.<ext>
    where i is the frame number. if no file found for i < N, raise error.
    otherwise, return found file
    """
    f, ext = os.path.splitext(fname)
    for i in range(N):
        fname = f + f"{i:04d}" + ext
        if os.path.isfile(fname):
            return fname

    raise FileNotFoundError(f"File {fname} not found")


def _all_anim_frames(fname, N=1000):
    """Search for filename in the format
    <main_fname><i:04d>.<ext>
    where i is the frame number. if no file found for i < N, raise error.
    otherwise, return found file
    """
    f, ext = os.path.splitext(fname)
    for i in range(N):
        fname = f + f"{i:04d}" + ext
        if os.path.isfile(fname):
            yield fname
        else:
            break


def remove_ext(fname):
    return os.path.splitext(fname)[0]  # remove extension if given


class Compositor:
    """Compositor output - for handling file outputs, and managing Compositor node tree"""

    def __init__(
        self,
        view_layer="ViewLayer",
        background_color: tuple = None,
        rgb_color_space: str = None,
    ):
        """
        :param view_layer: Name of View Layer to render
        :param background_color: If given, RGB[A] tuple in range [0-1], will overwrite World background with solid color (while retaining lighting effects).
        :param rgb_color_space: Color transform for RGB only. Defaults to `AgX Base sRGB` for Blender 4+, and `Filmic sRGB` for older versions.
        """

        if rgb_color_space is None:
            rgb_color_space = _default_color_space()

        # Create compositor node tree
        bpy.context.scene.use_nodes = True
        self.node_tree = bpy.context.scene.node_tree

        self.view_layer = view_layer  #

        self.file_output_nodes = {}  # Mapping of output name to FileOutputNode
        self.mask_nodes = {}  # Mapping of mask pass index to CompositorNodeGroup
        self.overlays = {}
        self.aovs = []  # List of AOVs (used to update before rendering)

        # We set no view transform by default - all color correction
        # to be handled in compositor, so that AOVs are in raw color space
        # Handling different on different devices
        for (dd, vt) in [('None', 'Standard'), ('sRGB', 'Standard')]:
            try:
                bpy.context.scene.display_settings.display_device = dd
                bpy.context.scene.view_settings.view_transform = vt
                break
            except TypeError:
                continue
        else:
            raise ValueError("No acceptable display devices found - please report this issue on GitHub.")

        # Handled more simply in 4.2+:
        if version.is_version_plus(4.2):
            bpy.context.scene.view_settings.view_transform = 'Raw'

        # Socket to be used as RGB input for anything. Defined separately in case of applying overlays (e.g. background color)
        self._rgb_socket = get_node_by_name(self.node_tree, "Render Layers").outputs[
            "Image"
        ]
        self._set_rgb_color_space(rgb_color_space)
        if background_color is not None:
            self._set_background_color(background_color)

    def _tidy_tree(self):
        """Tidy up node tree"""
        tidy_tree(self.node_tree)

    @property
    def render_layers_node(self):
        return get_node_by_name(self.node_tree, "Render Layers")

    def _get_render_layer_output(self, key: str):
        """Get output socket from Render Layers node"""
        if key == "Image":  # special case
            return self._rgb_socket
        else:
            return self.render_layers_node.outputs[key]

    def get_mask(
        self, index, input_rgb: Union[str, CompositorNodeGroup], anti_aliasing=False
    ) -> CompositorNodeGroup:
        """Get mask node from pass index. If not found, create new mask node"""
        bpy.context.scene.view_layers[
            self.view_layer
        ].use_pass_object_index = True  # Make sure object index is enabled

        if index not in self.mask_nodes:
            if isinstance(input_rgb, str):
                ip_node = self._get_render_layer_output(input_rgb)

            elif isinstance(input_rgb, CompositorNodeGroup):
                ip_node = input_rgb.outputs["Image"]

            else:
                raise TypeError(
                    f"input_rgb must be str or CompositorNodeGroup, got {type(input_rgb)}"
                )

            dtype = (
                "Float" if isinstance(ip_node, bpy.types.NodeSocketFloat) else "Color"
            )

            cng = MaskOverlay(
                f"Mask - ID: {index} - Input {input_rgb}",
                self.node_tree,
                index=index,
                dtype=dtype,
                use_antialiasing=anti_aliasing,
            )

            self.node_tree.links.new(ip_node, cng.input("Image"))
            self.node_tree.links.new(
                self._get_render_layer_output("IndexOB"), cng.input("IndexOB")
            )
            self.mask_nodes[index] = cng

        self._tidy_tree()
        return self.mask_nodes[index]

    def get_bounding_box_visual(
        self, col=(0.0, 0.0, 255.0, 255.0), thickness: int = 5
    ) -> BoundingBoxOverlay:
        """
        Return a bounding box visual overlay.

        :param col: (3,) or (N, 3) Color(s) of bounding box(es) [in BGR]
        :param thickness: (,) or (N,) Thickness(es) of bounding box(es)
        """

        cng = BoundingBoxOverlay(
            f"Bounding Box Visual", self.node_tree, col=col, thickness=thickness
        )
        self.node_tree.links.new(
            self._get_render_layer_output("Image"), cng.input("Image")
        )

        if "bbox" in self.overlays:
            raise ValueError(
                "Only allowed one BBox overlay (it can contain multiple objects)."
            )

        self.overlays["bbox"] = cng

        self._tidy_tree()
        return cng

    def get_keypoints_visual(
        self,
        marker: str = "x",
        color: tuple = (0, 0, 255),
        size: int = 5,
        thickness: int = 2,
    ) -> KeypointsOverlay:
        """
        Return a keypoints visual overlay.

        :param marker: Marker type, either [c/circle], [s/square], [t/triangle] or [x]. Default 'x'
        :param size: Size of marker. Default 5
        :param color: Color of marker, RGB or RGBA, default (0, 0, 255) (red)
        :param thickness: Thickness of marker. Default 2
        """

        cng = KeypointsOverlay(
            f"Keypoints Visual",
            self.node_tree,
            marker=marker,
            color=color,
            size=size,
            thickness=thickness,
        )
        self.node_tree.links.new(
            self._get_render_layer_output("Image"), cng.input("Image")
        )

        if "keypoints" in self.overlays:
            raise ValueError("Only allowed one Keypoints overlay.")

        self.overlays["keypoints"] = cng

        self._tidy_tree()
        return cng

    def get_axes_visual(self, size: int = 1, thickness: int = 2) -> AxesOverlay:
        """
        Return an axes visual overlay.

        :param size: Size of axes. Default 100
        :param thickness: Thickness of axes. Default 2
        """

        cng = AxesOverlay(
            f"Axes Visual", self.node_tree, size=size, thickness=thickness
        )
        self.node_tree.links.new(
            self._get_render_layer_output("Image"), cng.input("Image")
        )

        if "axes" in self.overlays:
            raise ValueError("Only allowed one Axes overlay.")

        self.overlays["axes"] = cng
        self._tidy_tree()
        return cng

    def stack_visuals(self, *visuals: AlphaImageOverlay) -> AlphaImageOverlay:
        """Given a series of image overlays, stack them and return to be used as a single output node.

        :param *visuals: Stack of overlays to add."""

        if len(visuals) < 2:
            raise ValueError("Requires at least 2 visuals to stack")

        # No need to store these overlays separately in self.overlays, but need to check they're all present
        for overlay in visuals:
            if overlay not in self.overlays.values():
                raise ValueError(
                    f"Visual {overlay} not found in Compositor. Make sure it was obtained via the Compositor."
                )

        # Stack the output of the previous to the input of the next
        for va, vb in zip(visuals, visuals[1:]):
            self.node_tree.links.new(va.output("Image"), vb.input("Image"))

        return visuals[-1]

    def get_depth_visual(
        self, max_depth=1, col: tuple = (255, 255, 255)
    ) -> CompositorNodeGroup:
        """Get depth visual, which normalizes depth values so max_depth = col,
        and any values below that are depth/max_depth * col.

        :param max_depth: Maximum depth value to normalize to
        :param col: Color of maximum depth value. 0-255 RGB or RGBA."""

        if "Depth" not in self.render_layers_node.outputs:
            render_depth()

        # convert col to 0-1, RGBA
        col = ([i / 255 for i in col] + [1])[:4]

        cng = DepthVis(self.node_tree, max_depth=max_depth, col=col)
        self.node_tree.links.new(
            self._get_render_layer_output("Depth"), cng.input("Depth")
        )

        self._tidy_tree()
        return cng

    def define_output(
        self,
        input_data: Union[str, CompositorNodeGroup, AOV],
        directory: str = ".",
        file_name: str = None,
        file_format: str = "PNG",
        color_mode: str = "RGBA",
        jpeg_quality: int = 90,
        png_compression: int = 15,
        color_depth: str = "8",
        EXR_color_depth: str = "32",
        name: str = None,
    ) -> str:
        """Add a connection between a valid render output, and a file output node.

        This should only be called once per output (NOT inside a loop).
        Inside the loop, only call :attr:`~update_filename`, :attr:`update_all_filenames` :attr:`~update_directory`

        All outputs will be defined in raw color space (no color correction), except for the RGB output,
        and any overlays on this output (e.g. Bounding Box Visualization)

        :param input_data: If :class:`str`,  will get the input_data from that key in the render_layers_node. If :class:`~CompositorNodeGroup`, use that node as input. If :class:`AOV`, use that AOV as input (storing AOV).
        :param directory: Directory to save output to
        :param file_name: Name of file to save output to. If None, will use `name` (or `input_data` if `name` is None)
        :param file_format: File format to save output as. Must be in :class:`AVAILABLE_FORMATS`
        :param color_mode: Color mode to save output as.
        :param jpeg_quality: Quality of JPEG output.
        :param png_compression: Compression of PNG output.
        :param color_depth: Color depth of output.
        :param EXR_color_depth: Color depth of EXR output.
        :param name: Name of output. If not given, will take the str representation of input_data

        :returns: Name of output, which can be used to update filename or directory

        """

        if isinstance(input_data, AOV):
            self.aovs.append(input_data)
            input_data = (
                input_data.name
            )  # name is sufficient to pull from render_layers_node

        assert (
            file_format in format_to_extension
        ), f"File format `{file_format}` not supported. Options are: {list(format_to_extension.keys())}"

        if name is None:
            name = str(input_data)

        node_name = f"File Output {name}"

        # check node doesn't exist
        if name in self.file_output_nodes:
            raise ValueError(
                f"File output `{name}` already exists. Only call define_output once per output type."
            )

        if file_name is None:  # if fname is not given, use input_name
            file_name = name

        file_name = remove_ext(file_name)
        directory = os.path.abspath(directory)  # make sure directory is absolute
        os.makedirs(directory, exist_ok=True)

        # Create new 'File Output' node in compositor
        node = self.node_tree.nodes.new("CompositorNodeOutputFile")
        node.name = node_name

        if isinstance(input_data, str):
            self.node_tree.links.new(
                self._get_render_layer_output(input_data), node.inputs["Image"]
            )

        elif isinstance(input_data, CompositorNodeGroup):  # add overlay in between
            self.node_tree.links.new(input_data.outputs[0], node.inputs["Image"])

        else:
            raise NotImplementedError(
                f"input_data must be either str or CompositorNodeGroup, got {type(input_data)}"
            )

        # Set file output node properties
        node.label = node_name
        node.base_path = directory
        node.file_slots[0].path = file_name

        # File format kwargs
        node.format.file_format = file_format
        node.format.color_mode = color_mode
        node.format.quality = jpeg_quality
        node.format.compression = png_compression
        node.format.color_depth = (
            color_depth if file_format != "OPEN_EXR" else EXR_color_depth
        )

        self.file_output_nodes[name] = node

        self._tidy_tree()
        return name

    def update_filename(self, key: str, fname: str):
        """Reassign the filename (not directory) for a given file output node.

        :param key: key of output, as given in `define_output`
        :param fname: new filename, without extension"""
        fname = remove_ext(fname)
        node = self.file_output_nodes[str(key)]
        node.file_slots[0].path = fname

    def update_all_filenames(self, fname: str):
        """Reassign all filenames (not directories) for all file output nodes.

        :param fname: new filename, without extension"""
        fname = remove_ext(fname)
        for node in self.file_output_nodes.values():
            node.file_slots[0].path = fname

    def update_directory(self, key: str, directory: str):
        """Reassign the directory for a given file output node

        :param key: key of output, as given in `define_output`
        :param directory: new directory
        """
        node = self.file_output_nodes[str(key)]
        node.base_path = directory

    def _fix_namings_static(self, prefix=None):
        """After rendering a static scene,
        File Output node has property that frame number gets added to filename.
        Fix that here"""

        prefix = "" if prefix is None else prefix + "_"

        for node in self.file_output_nodes.values():
            # get expected file name and extension
            ext = format_to_extension[node.format.file_format]
            bad_file_name = _get_badfname(
                os.path.join(node.base_path, node.file_slots[0].path + ext)
            )
            target_file_name = os.path.join(
                node.base_path, prefix + node.file_slots[0].path + ext
            )
            shutil.move(bad_file_name, target_file_name)

    def _fix_namings_animation(self, prefix=None):
        """After rendering an animation, find all the files that have been rendered,
        and prepend prefix to them"""
        if prefix is None:
            return

        for node in self.file_output_nodes.values():
            # get expected file name and extension
            ext = format_to_extension[node.format.file_format]
            for fname in _all_anim_frames(
                os.path.join(node.base_path, node.file_slots[0].path + ext)
            ):
                target_file_name = os.path.join(
                    node.base_path, prefix + "_" + os.path.basename(fname)
                )
                shutil.move(fname, target_file_name)

    def _update_aovs(self):
        """Update any AOVs that are connected to the render layers node"""
        for aov in self.aovs:
            aov.update()

    def render(
        self,
        camera: Union[Camera, List[Camera]] = None,
        scene: bpy.types.Scene = None,
        annotations: AnnotationHandler = None,
        animation: bool = False,
        frame_start: int = 0,
        frame_end: int = 250,
    ):
        """Render the scene.

        :param camera: Camera(s) to render from. If None, will use `scene.camera`. If multiple, will render each camera separately, appending the camera's names as the output file names.
        :param scene: Scene to render. If None, will use `bpy.context.scene`.
        :param annotations: Object containing annotation information for each camera view to be used for overlays
        :param animation: If True, will render an animation, using `frame_start` and `frame_end` as the start and end frames.
        :param frame_start: Start frame for animation.
        :param frame_end: End frame for animation.
        """

        if scene is None:
            scene = bpy.context.scene

        _original_active_camera = bpy.context.scene.camera
        if camera is None:
            camera = Camera()

        if animation:
            scene.frame_start = frame_start
            scene.frame_end = frame_end

        multi_camera = isinstance(camera, list)
        if not multi_camera:
            camera = [camera]

        for cam in camera:
            if annotations is not None:
                annotation = annotations.get_annotation_by_camera(cam.name)

                # apply overlays PER CAMERA
                for oname, overlay in self.overlays.items():
                    overlay.update(
                        getattr(annotation, oname), camera=cam, scene=scene
                    )  # multi kwargs

            self._update_aovs()

            scene.camera = cam.obj
            render(animation=animation)

            if animation:
                self._fix_namings_animation(prefix=cam.name if multi_camera else None)

            else:
                self._fix_namings_static(prefix=cam.name if multi_camera else None)

        # reset active camera
        scene.camera = _original_active_camera

    def _set_rgb_color_space(self, color_space: str = "Filmic sRGB"):
        """Color spaces are all handled manually within compositor (so that we keep
        AOVs in raw space). So set the color space for RGB socket here."""

        color_space_node = self.node_tree.nodes.new("CompositorNodeConvertColorSpace")

        for key in ["Linear", "Non-Color"]:
            try:
                color_space_node.from_color_space = key
                break
            except TypeError:
                continue
        else:
            raise ValueError("No acceptable color spaces found - please report this issue on GitHub.")

        color_space_node.to_color_space = color_space

        self.node_tree.links.new(self._rgb_socket, color_space_node.inputs[0])
        self._rgb_socket = color_space_node.outputs[0]

    def _set_background_color(self, color: tuple = (0, 0, 0)):
        """Set a solid background color, instead of transparent.
        Will remove the visuals of existing world background (but not the lighting effects).

        :param color: RGBA color, in range [0, 1]
        """

        world.set_transparent()

        rgba = color
        if len(rgba) == 3:
            rgba = (*rgba, 1)

        rgb_node = self.node_tree.nodes.new("CompositorNodeRGB")
        rgb_node.outputs[0].default_value = rgba

        mix_node = self.node_tree.nodes.new("CompositorNodeMixRGB")

        self.node_tree.links.new(self._rgb_socket, mix_node.inputs[2])
        self.node_tree.links.new(rgb_node.outputs[0], mix_node.inputs[1])

        self.node_tree.links.new(
            self._get_render_layer_output("Alpha"), mix_node.inputs["Fac"]
        )
        self._rgb_socket = mix_node.outputs["Image"]



"""Utilities for neatly arranging nodes in the node editor,
to make debugging easier."""

import bpy
from collections import defaultdict
from copy import copy


def get_source_nodes(node):
    for i in node.inputs:
        for link in i.links:
            yield link.from_node


def get_sink_nodes(node):
    for o in node.outputs:
        for link in o.links:
            yield link.to_node


def split_to_islands(nodes):
    """Given a node list, return a list of sets, each one being a fully disconnected island"""

    def dfs(node, island):
        """Depth-first search from node."""
        node["visited"] = True
        islands[node] = island
        for linked_node in [*get_source_nodes(node), *get_sink_nodes(node)]:
            if not linked_node["visited"]:
                dfs(linked_node, island)

    islands = {}
    island_counter = 0

    # Set all nodes to unvisited
    for node in nodes:
        node["visited"] = False

    # Start a DFS from each unvisited node
    for node in nodes:
        if not node["visited"]:
            dfs(node, island_counter)
            island_counter += 1

    # Group nodes by island
    grouped_islands = []
    for i in range(island_counter):
        grouped_islands.append({k for k, v in islands.items() if v == i})

    return grouped_islands


def calc_depth(node_island):
    """Given a node island (i.e. all nodes in the same fully connected component),
    set the depth of each node to be the maximum depth of all its input nodes + 1.
    Once complete, normalize these depths so the first item is 0"""

    for nodes in node_island:
        nodes["depth"] = None

    node_island_copy = copy(node_island)
    start_node = node_island_copy.pop()
    start_node["depth"] = 0
    nodes_to_use = {start_node}

    # First Pass - assign a depth to each node by traversing the graph from the start node
    while any(node["depth"] is None for node in node_island):
        node = nodes_to_use.pop()
        for sink_node in get_sink_nodes(node):
            if sink_node in node_island:
                if sink_node["depth"] is None:
                    sink_node["depth"] = node["depth"] + 1
                    nodes_to_use.add(sink_node)

        for source_node in get_source_nodes(node):
            if source_node in node_island:
                if source_node["depth"] is None:
                    source_node["depth"] = node["depth"] - 1
                    nodes_to_use.add(source_node)

    # Second Pass - set each depth to be the maximum depth of all its input nodes + 1
    depth_changed = True
    while depth_changed:
        depth_changed = False
        for node in node_island:
            sources = [s for s in get_source_nodes(node) if s in node_island]
            if sources:
                new_depth = max(node["depth"] for node in sources) + 1
                if node["depth"] != new_depth:
                    node["depth"] = new_depth
                    depth_changed = True

    # Normalize depths
    min_depth = min(node["depth"] for node in node_island)
    for node in node_island:
        node["depth"] -= min_depth

    return node_island


def tidy_tree(node_tree: bpy.types.NodeTree, dX: int = 400, dY: int = 200):
    """Search through tree, positioning nodes in a grid based on their depth and connectivity.

    :param node_tree: node tree to tidy
    :param dX: horizontal distance between nodes
    :param dY: vertical distance between nodes"""

    nodes = node_tree.nodes
    islands = split_to_islands(nodes)

    y = 0  # track running height to manage multiple islands

    height = defaultdict(int)  # track height of each depth level
    for island in islands:
        island = calc_depth(island)

        # want to center each depth level, so set heights accordingly
        for i in range(max(node["depth"] for node in island) + 1):
            # Adjusting the initial height calculation to account for node sizes.
            depth_nodes = [node for node in island if node["depth"] == i]
            total_height_for_depth = sum(node.dimensions.y for node in depth_nodes)
            spacing_needed = dY * (len(depth_nodes) - 1)
            height[i] = -(total_height_for_depth + spacing_needed) / 2

        for node in island:
            node.location = (node["depth"] * dX, y + height[node["depth"]])
            height[node["depth"]] += (
                node.dimensions.y + dY
            )  # Incrementing by the node's height

        y += max(height.values()) + dY



import bpy
from .utils import project_points
import mathutils
import numpy as np
from ..blender.mesh import Mesh
from ..blender.camera import Camera
from .annotation_handler import AnnotationHandler
from typing import List, Union

BBOX_FMTS = ["x1y1x2y2", "xywh"]


def _bounding_box(
    object: Mesh,
    camera: bpy.types.Camera = None,
    scene: bpy.types.Scene = None,
    return_fmt: str = "x1y1x2y2",
    normalized: bool = False,
    invert_y: bool = True,
) -> tuple:
    """Get the bounding box of an object in camera space.
    Achieve this by projecting all vertices of the object to the image plane,
    and taking the min/max of the resulting coordinates.

    :param object: Mesh
    :param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
    :param scene: bpy.types.Scene (if None, use bpy.context.scene)
    :param return_fmt: one of ['x1y1x2y2', 'xywh']
    :param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
    """

    if camera is None:
        camera = bpy.context.scene.camera

    if scene is None:
        scene = bpy.context.scene

    verts = object._get_all_vertices("WORLD")
    coords_2d = project_points(verts, scene, camera, invert_y=invert_y)

    if normalized:
        coords_2d[:, 0] /= scene.render.resolution_x
        coords_2d[:, 1] /= scene.render.resolution_y

    xmin, ymin = coords_2d.min(axis=0)
    xmax, ymax = coords_2d.max(axis=0)

    if return_fmt == "x1y1x2y2":
        return xmin, ymin, xmax, ymax

    elif return_fmt == "xywh":
        return xmin, ymin, xmax - xmin, ymax - ymin

    else:
        raise ValueError(
            f"Invalid return_fmt: {return_fmt}. Must be one of {BBOX_FMTS}"
        )


def _bounding_boxes(
    objects: List[Mesh],
    camera: bpy.types.Camera = None,
    scene: bpy.types.Scene = None,
    return_fmt: str = "x1y1x2y2",
    normalized: bool = False,
    invert_y: bool = True,
) -> List[tuple]:
    """Run `bounding_box` for multiple objects

    :param objects: List of Mesh objects
    :param camera: bpy.types.Camera (if None, use bpy.context.scene.camera)
    :param scene: bpy.types.Scene (if None, use bpy.context.scene)
    :param return_fmt: one of ['x1y1x2y2', 'xywh']
    :param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
    """

    return [
        _bounding_box(
            obj,
            camera=camera,
            scene=scene,
            return_fmt=return_fmt,
            normalized=normalized,
            invert_y=invert_y,
        )
        for obj in objects
    ]


def bounding_boxes(
    objects: List[Mesh],
    camera: Union[Camera, List[Camera]] = None,
    scene: bpy.types.Scene = None,
    return_fmt: str = "x1y1x2y2",
    normalized: bool = False,
    invert_y: bool = True,
) -> AnnotationHandler:
    """
    Project bounding boxes of all objects in list to all cameras provided.
    :param objects: List of Mesh objects
    :param camera: Either a single :class:`~blendersynth.blender.camera.Camera` object, or a list of them. If None, use bpy.context.scene.camera
    :param scene: bpy.types.Scene (if None, use bpy.context.scene)
    :param return_fmt: one of ['x1y1x2y2', 'xywh']
    :param normalized: if True, return normalized coordinates (0-1) instead of pixel coordinates
    :param invert_y: if True, y is measured from the top of the image, otherwise from the bottom (Blender measures from bottom)
    """

    return_dict = {}

    if camera is None:
        camera = bpy.context.scene.camera

    if not isinstance(camera, list):
        camera = [camera]

    for cam in camera:
        return_dict[cam.name] = _bounding_boxes(
            objects,
            camera=cam,
            scene=scene,
            return_fmt=return_fmt,
            normalized=normalized,
            invert_y=invert_y,
        )

    return AnnotationHandler.from_annotations(return_dict, ann_type="bbox")



import json
import os
from typing import Union


def save_label(data: Union[dict, list], pth: str):
    """Save a label to a file.

    :param data: The data to save.
    :param pth: The path to save the data to."""

    os.makedirs(os.path.dirname(pth), exist_ok=True)
    with open(pth, "w") as outfile:
        json.dump(data, outfile)



import bpy
from .utils import project_points
import numpy as np
from .annotation_handler import AnnotationHandler
from ..blender.camera import Camera
from typing import List, Union


def _project_keypoints(
    points3d: np.ndarray, camera: bpy.types.Camera = None, scene: bpy.types.Scene = None
) -> np.ndarray:
    """Project 3D world points to 2D image coordinates.

    Projects using :func:`blendersynth.annotations.utils.project_points`.

    :param points3d: Nx3 array of 3D points
    :param scene: scene to use (if None, `use bpy.context.scene`)
    :param camera: camera to project through (if None, use `bpy.context.scene.camera`)

    :return: Nx2 array of 2D image coordinates
    """

    if scene is None:
        scene = bpy.context.scene

    if camera is None:
        camera = scene.camera

    # project points
    return project_points(points3d, scene, camera)


def project_keypoints(
    points3d: np.ndarray,
    camera: Union[Camera, List[Camera]] = None,
    scene: bpy.types.Scene = None,
) -> AnnotationHandler:
    """Project 3D world points to 2D image coordinates for each camera.

    :param points3d: Nx3 array of 3D points
    :param camera: Either a single :class:`~blendersynth.blender.camera.Camera` object, or a list of them. If None, use bpy.context.scene.camera
    :param scene: bpy.types.Scene (if None, use bpy.context.scene)
    """

    return_dict = {}

    if camera is None:
        camera = bpy.context.scene.camera

    if not isinstance(camera, list):
        camera = [camera]

    for cam in camera:
        return_dict[cam.name] = _project_keypoints(points3d, camera=cam, scene=scene)

    return AnnotationHandler.from_annotations(return_dict, ann_type="keypoints")
